import { ServiceError } from "nats";
import Adaptor from "./adaptors/adaptor.ts";
import { RequestError } from "./error.ts";
import { Message, Request, Response } from "./messages.ts";
import { RouteSubscribeTypeFn } from "./types.ts";
import { z } from "zod";

export default class Service {
  private name: string;

  private adaptors: Record<string, Adaptor>;

  constructor(name: string) {
    this.name = name;
    this.adaptors = {};
  }

  public addAdaptor(name: string, adaptor: Adaptor) {
    this.adaptors[name] = adaptor;
  }

  public subscribe<T extends z.ZodType<O>, O, U>(
    adaptor: string,
    subject: string,
    fn: RouteSubscribeTypeFn<z.infer<T>, U>,
    schema: T | undefined = undefined,
  ) {
    this.adaptors[adaptor].subscribe(
      `${this.name}.${subject}`,
      async (rawReq) => {
        const msg: Request<T> = JSON.parse(rawReq);

        const message: Message<z.infer<T>> = { from: this.name };

        if (msg.data && schema) {
          const validate = schema.safeParse(msg.data);
          if (!validate.success) {
            return JSON.stringify(
              {
                statusCode: 400,
                data: validate.error.toString(),
              } satisfies Response<string>,
            );
          } else {
            message.data = validate.data;
          }
        }

        let res;
        try {
          res = await fn(message);
        } catch (err) {
          if (err instanceof ServiceError) {
            res = { data: err.message, statusCode: err.code } as Response<
              string
            >;
          } else {
            res = { data: "unknow error append", statusCode: 500 } as Response<
              string
            >;
          }
        }

        return JSON.stringify(res);
      },
    );
  }

  public async request<T, O, U extends z.ZodType<O>>(
    adaptor: string,
    req: Request<T>,
    schema?: z.ZodType<O>,
  ): Promise<Response<z.infer<U>>> {
    if (!this.adaptors[adaptor]) {
      throw new Error(`${adaptor} adaptor not exist`);
    }

    const rawReq = JSON.stringify(req);

    try {
      const rawRes = await this.adaptors[adaptor].request(
        `${req.service}.${req.subject}`,
        rawReq,
      );
      const res: Response<z.infer<U>> = JSON.parse(rawRes);

      if (res.statusCode < 200 || res.statusCode >= 299) {
        throw new RequestError("error while request", res.statusCode);
      }

      if (res.data && schema) {
        const validate = schema.safeParse(res.data);

        if (!validate.success) {
          throw new ServiceError(400, validate.error.toString());
        } else {
          res.data = validate.data;
        }
      }

      return res;
    } catch (err) {
      if (err instanceof RequestError) {
        throw err;
      }

      throw new RequestError("unexpected error", 500);
    }
  }

  public async listen() {
    for (const index in this.adaptors) {
      await this.adaptors[index].listen(this.name);
    }
  }

  public async stop() {
    for (const index in this.adaptors) {
      await this.adaptors[index].stop();
    }
  }
}
